@@ -25,6 +25,7 @@
 
 #include <QtCore/QCoreApplication>
 #include <QtCore/QDateTime>
+#include <QtCore/QDir>
 #include <QtCore/QFile>
 #include <QtCore/QFileInfo>
 #include <QtCore/QHash>
@@ -57,20 +58,20 @@ class AutoMoc
 
     private:
         void dotFilesCheck(bool);
+        void lazyInitMocDefinitions();
         void lazyInit();
         bool touch(const QString &filename);
         bool generateMoc(const QString &sourceFile, const QString &mocFileName);
-        void waitForProcesses();
         void printUsage(const QString &);
         void printVersion();
         void echoColor(const QString &msg)
         {
-            QProcess *cmakeEcho = new QProcess;
-            cmakeEcho->setProcessChannelMode(QProcess::ForwardedChannels);
+            QProcess cmakeEcho;
+            cmakeEcho.setProcessChannelMode(QProcess::ForwardedChannels);
             QStringList args(cmakeEchoColorArgs);
             args << msg;
-            cmakeEcho->start(cmakeExecutable, args, QIODevice::NotOpen);
-            processes.enqueue(Process(cmakeEcho, QString()));
+            cmakeEcho.start(cmakeExecutable, args, QIODevice::NotOpen);
+            cmakeEcho.waitForFinished(-1);
         }
 
         QString builddir;
@@ -83,13 +84,6 @@ class AutoMoc
         const bool verbose;
         QTextStream cerr;
         QTextStream cout;
-        struct Process
-        {
-            Process(QProcess *a, const QString &b) : qproc(a), mocFilePath(b) {}
-            QProcess *qproc;
-            QString mocFilePath;
-        };
-        QQueue<Process> processes;
         bool failed;
         bool automocCppChanged;
         bool generateAll;
@@ -125,7 +119,7 @@ int main(int argc, char **argv)
 
 AutoMoc::AutoMoc()
     : verbose(!qgetenv("VERBOSE").isEmpty()), cerr(stderr), cout(stdout), failed(false),
-    automocCppChanged(false), doTouch(false)
+    automocCppChanged(false), generateAll(false), doTouch(false)
 {
     const QByteArray colorEnv = qgetenv("COLOR");
     cmakeEchoColorArgs << QLatin1String("-E") << QLatin1String("cmake_echo_color") 
@@ -133,18 +127,13 @@ AutoMoc::AutoMoc()
         << QLatin1String("--bold");
 }
 
-void AutoMoc::lazyInit()
+void AutoMoc::lazyInitMocDefinitions()
 {
-    const QStringList &args = QCoreApplication::arguments();
-    mocExe = args[4];
-    cmakeExecutable = args[5];
-
-    if (args.size() > 6) {
-        if (args[6] == QLatin1String("--touch")) {
-            doTouch = true;
-        }
+    static bool done = false;
+    if (done) {
+        return;
     }
-
+    done = true;
     QByteArray line = dotFiles.readLine();
     dotFilesCheck(line == "MOC_COMPILE_DEFINITIONS:\n");
     line = dotFiles.readLine().trimmed();
@@ -166,8 +155,23 @@ void AutoMoc::lazyInit()
             }
         }
     }
+}
 
-    line = dotFiles.readLine();
+void AutoMoc::lazyInit()
+{
+    const QStringList &args = QCoreApplication::arguments();
+    mocExe = args[4];
+    cmakeExecutable = args[5];
+
+    if (args.size() > 6) {
+        if (args[6] == QLatin1String("--touch")) {
+            doTouch = true;
+        }
+    }
+
+    lazyInitMocDefinitions();
+
+    QByteArray line = dotFiles.readLine();
     dotFilesCheck(line == "MOC_INCLUDES:\n");
     line = dotFiles.readLine().trimmed();
     const QStringList &incPaths = QString::fromUtf8(line).split(';', QString::SkipEmptyParts);
@@ -248,8 +252,6 @@ bool AutoMoc::run()
         builddir += '/';
     }
 
-    generateAll = !outfile.exists();
-
     dotFiles.setFileName(args[1] + QLatin1String(".files"));
     dotFiles.open(QIODevice::ReadOnly | QIODevice::Text);
 
@@ -257,6 +259,20 @@ bool AutoMoc::run()
     dotFilesCheck(line == "SOURCES:\n");
     const QStringList &sourceFiles = QString::fromUtf8(dotFiles.readLine().trimmed()).split(';', QString::SkipEmptyParts);
 
+    if (outfile.exists()) {
+        // set generateAll = true if MOC_COMPILE_DEFINITIONS changed
+        outfile.open(QIODevice::ReadOnly | QIODevice::Text);
+        QByteArray buf = outfile.readLine();
+        // the second line contains the joined mocDefinitions
+        buf = outfile.readLine();
+        buf.chop(1); // remove trailing \n
+        lazyInitMocDefinitions();
+        generateAll = (buf != mocDefinitions.join(QString(QLatin1Char(' '))).toUtf8());
+        outfile.close();
+    } else {
+        generateAll = true;
+    }
+
     // the program goes through all .cpp files to see which moc files are included. It is not really
     // interesting how the moc file is named, but what file the moc is created from. Once a moc is
     // included the same moc may not be included in the _automoc.cpp file anymore. OTOH if there's a
@@ -265,7 +281,7 @@ bool AutoMoc::run()
     QHash<QString, QString> includedMocs;    // key = moc source filepath, value = moc output filepath
     QHash<QString, QString> notIncludedMocs; // key = moc source filepath, value = moc output filename
 
-    QRegExp mocIncludeRegExp(QLatin1String("[\n]\\s*#\\s*include\\s+[\"<](moc_[^ \">]+\\.cpp|[^ \">]+\\.moc)[\">]"));
+    QRegExp mocIncludeRegExp(QLatin1String("[\n]\\s*#\\s*include\\s+[\"<]((?:[^ \">]+/)?moc_[^ \">/]+\\.cpp|[^ \">]+\\.moc)[\">]"));
     QRegExp qObjectRegExp(QLatin1String("[\n]\\s*Q_OBJECT\\b"));
     QStringList headerExtensions;
 #ifdef Q_OS_WIN
@@ -362,15 +378,28 @@ bool AutoMoc::run()
                 do { // call this for every moc include in the file
                     const QString currentMoc = mocIncludeRegExp.cap(1);
                     //qDebug() << "found moc include: " << currentMoc << " at offset " << matchOffset;
-                    QString basename = QFileInfo(currentMoc).completeBaseName();
-                    const bool moc_style = currentMoc.startsWith("moc_");
+                    const QFileInfo currentMocInfo(currentMoc);
+                    QString basename = currentMocInfo.completeBaseName();
+                    const bool moc_style = basename.startsWith("moc_");
+
+                    // If the moc include is of the moc_foo.cpp style we expect the Q_OBJECT class
+                    // declaration in a header file.
+                    // If the moc include is of the foo.moc style we need to look for a Q_OBJECT
+                    // macro in the current source file, if it contains the macro we generate the
+                    // moc file from the source file, else from the header.
+                    //
+                    // TODO: currently any .moc file name will be used if the source contains
+                    // Q_OBJECT
                     if (moc_style || qObjectRegExp.indexIn(contentsString) < 0) {
                         if (moc_style) {
+                            // basename should be the part of the moc filename used for finding the
+                            // correct header, so we need to remove the moc_ part
                             basename = basename.right(basename.length() - 4);
                         }
+
                         bool headerFound = false;
                         foreach (const QString &ext, headerExtensions) {
-                            QString sourceFilePath = absPath + basename + ext;
+                            const QString &sourceFilePath = absPath + basename + ext;
                             if (QFile::exists(sourceFilePath)) {
                                 headerFound = true;
                                 includedMocs.insert(sourceFilePath, currentMoc);
@@ -379,11 +408,34 @@ bool AutoMoc::run()
                             }
                         }
                         if (!headerFound) {
+                            // the moc file is in a subdir => look for the header in the same subdir
+                            if (currentMoc.indexOf('/') != -1) {
+                                const QString &filepath = absPath + currentMocInfo.path() + QLatin1Char('/') + basename;
+
+                                foreach (const QString &ext, headerExtensions) {
+                                    const QString &sourceFilePath = filepath + ext;
+                                    if (QFile::exists(sourceFilePath)) {
+                                        headerFound = true;
+                                        includedMocs.insert(sourceFilePath, currentMoc);
+                                        notIncludedMocs.remove(sourceFilePath);
+                                        break;
+                                    }
+                                }
+                                if (!headerFound) {
+                                    cerr << "automoc4: The file \"" << absFilename <<
+                                        "\" includes the moc file \"" << currentMoc << "\", but neither \"" <<
+                                        absPath + basename + "{" + headerExtensions.join(",") + "}\" nor \"" <<
+                                        filepath + "{" + headerExtensions.join(",") + "}" <<
+                                        "\" exist." << endl;
+                                    ::exit(EXIT_FAILURE);
+                                }
+                            } else {
                                 cerr << "automoc4: The file \"" << absFilename <<
                                     "\" includes the moc file \"" << currentMoc << "\", but \"" <<
                                     absPath + basename + "{" + headerExtensions.join(",") + "}" <<
-                                    "\" do not exist." << endl;
+                                    "\" does not exist." << endl;
                                 ::exit(EXIT_FAILURE);
+                            }
                         }
                     } else {
                         includedMocs.insert(absFilename, currentMoc);
@@ -420,7 +472,8 @@ bool AutoMoc::run()
 
     QByteArray automocSource;
     QTextStream outStream(&automocSource, QIODevice::WriteOnly);
-    outStream << "/* This file is autogenerated, do not edit */\n";
+    outStream << "/* This file is autogenerated, do not edit\n"
+        << mocDefinitions.join(QString(QLatin1Char(' '))) << "\n*/\n";
 
     if (notIncludedMocs.isEmpty()) {
         outStream << "enum some_compilers { need_more_than_nothing };\n";
@@ -436,9 +489,6 @@ bool AutoMoc::run()
         }
     }
 
-    // let all remaining moc processes finish
-    waitForProcesses();
-
     if (failed) {
         // if any moc process failed we don't want to touch the _automoc.cpp file so that
         // automoc4 is rerun until the issue is fixed
@@ -496,32 +546,18 @@ bool AutoMoc::touch(const QString &_filename)
     return true;
 }
 
-void AutoMoc::waitForProcesses()
-{
-    while (!processes.isEmpty()) {
-        Process proc = processes.dequeue();
-
-        bool result = proc.qproc->waitForFinished(-1);
-        //ignore errors from the cmake echo process
-        if (!proc.mocFilePath.isEmpty()) {
-            if (!result || proc.qproc->exitCode()) {
-                cerr << "automoc4: process for " << proc.mocFilePath
-                     << " failed: " << proc.qproc->errorString() << endl;
-                cerr << "pid to wait for: " << proc.qproc->pid() << endl;
-                cerr << "processes in queue: " << processes.size() << endl;
-                failed = true;
-                QFile::remove(proc.mocFilePath);
-            }
-        }
-        delete proc.qproc;
-    }
-}
-
 bool AutoMoc::generateMoc(const QString &sourceFile, const QString &mocFileName)
 {
     //qDebug() << Q_FUNC_INFO << sourceFile << mocFileName;
     const QString mocFilePath = builddir + mocFileName;
-    if (generateAll || QFileInfo(mocFilePath).lastModified() <= QFileInfo(sourceFile).lastModified()) {
+    QFileInfo mocInfo(mocFilePath);
+    if (generateAll || mocInfo.lastModified() <= QFileInfo(sourceFile).lastModified()) {
+        QDir mocDir = mocInfo.dir();
+        // make sure the directory for the resulting moc file exists
+        if (!mocDir.exists()) {
+            mocDir.mkpath(mocDir.path());
+        }
+
         static bool initialized = false;
         if (!initialized) {
             initialized = true;
@@ -533,19 +569,8 @@ bool AutoMoc::generateMoc(const QString &sourceFile, const QString &mocFileName)
             echoColor("Generating " + mocFileName);
         }
 
-        // we don't want too many child processes
-#ifdef Q_OS_FREEBSD
-    static const int max_processes = 0;
-#else
-    static const int max_processes = 10;
-#endif
-
-        if (processes.size() > max_processes) {
-            waitForProcesses();
-        }
-
-        QProcess *mocProc = new QProcess;
-        mocProc->setProcessChannelMode(QProcess::ForwardedChannels);
+        QProcess mocProc;
+        mocProc.setProcessChannelMode(QProcess::ForwardedChannels);
         QStringList args(mocIncludes + mocDefinitions);
 #ifdef Q_OS_WIN
         args << "-DWIN32";
@@ -555,15 +580,21 @@ bool AutoMoc::generateMoc(const QString &sourceFile, const QString &mocFileName)
         if (verbose) {
             cout << mocExe << " " << args.join(QLatin1String(" ")) << endl;
         }
-        mocProc->start(mocExe, args, QIODevice::NotOpen);
-        if (mocProc->waitForStarted()) {
-            processes.enqueue(Process(mocProc, mocFilePath));
+        mocProc.start(mocExe, args, QIODevice::NotOpen);
+        if (mocProc.waitForStarted()) {
+            const bool result = mocProc.waitForFinished(-1);
+            if (!result || mocProc.exitCode()) {
+                cerr << "automoc4: process for " << mocFilePath
+                     << " failed: " << mocProc.errorString() << endl;
+                cerr << "pid to wait for: " << mocProc.pid() << endl;
+                failed = true;
+                QFile::remove(mocFilePath);
+            }
             return true;
         } else {
             cerr << "automoc4: process for " << mocFilePath << "failed to start: " 
-                 << mocProc->errorString() << endl;
+                 << mocProc.errorString() << endl;
             failed = true;
-            delete mocProc;
         }
     }
     return false;
